#include "p18f4539.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usart.h>

#define pop_size 10 
#define dna_size 11
#define divisor 3640 //(32,767/(pop_size-1))
//global vars
char mutation_rate=25; //%of bits to be randomly mutated
static far rom char metric[dna_size] = {104,101,108,108,111,95,119,111,114,108,100}; // "hello world" in ascii
char pop[pop_size][dna_size];
char pop_new[pop_size][dna_size];
unsigned int fitnesses[pop_size];
char goal=dna_size;
char str[7]; //variable for converting numbers to strings

//function protos
void delay (void);
void quit(void);
void evaluate (char index);

void delay (void)
{
	int i;
	for (i = 0; i < 1000; i++)
	;
}

void quit(void){
	while (1){
		;
	}
}

void evaluate (char index){
	char  i,fitness;
	//evaluate the given index in pop against our metric and return a fitness factor
	//tallying fitness here is easy, just count the number of correct letters
	fitness=0;
	for (i=0;i<dna_size;i++){
		if (pop[index][i] == metric[i]) fitness++;
	}
	fitnesses[index]=fitness;
	return;
}

void main (void)
{
	char  i,j,l,n,c_index,c_index2,b,record,fittest;
	float sum,temp;
	unsigned int value;
	unsigned long count=0;
    // configure USART for 9600 baud
	OpenUSART( USART_TX_INT_OFF &
	USART_RX_INT_OFF &
	USART_ASYNCH_MODE &
	USART_EIGHT_BIT &
	USART_CONT_RX &
	USART_BRGH_LOW,31 );

	WriteUSART(254);//lcd cmd
	WriteUSART(1);//clear display

	//get current value of timer 1
	value = TMR1H<<8;
	value+= TMR1L;
	srand(value);

    //1. Create an initial population of candidate solutions
   //randomly initialize pop
	for (i=0;i<pop_size;i++){
		for (j=0;j<dna_size;j++){
			pop[i][j] = (rand()/1213)+95; //generate random number b/t 0 and 32,767
			//we need to scale down to an interval of 95-122 --> 27 numbers so divide by 1213 and shift up 95
		}
	}
    

   //2. Evaluate the initial population
	for (i=0;i<pop_size;i++){
		evaluate(i);
		itoa(fitnesses[i], str ); // convert to string
	}
   //check for success
	for (i=0;i<pop_size;i++){
		if (fitnesses[i]>=goal){
			WriteUSART(254);
			WriteUSART(128); //set cursor to 0 position
			for (b=0;b<dna_size;b++){
				putc(pop[i][b],stdout); //print ascii
				delay();
				delay();
			}
			quit(); //enter infinite loop
		}
	}
    
    //3. While no member of the population meets the criteria for success enter breeding loop
	while (1){
	    //a. select individuals into a mating pool-- proportional style
	    //b. Create a new population using crossover and mutation
		//find the fittest - they will be copied to the new generation
		record = 0;
		fittest = 0;
    	for (i = 0; i < pop_size; i++) {
      		if (fitnesses[i] > record) {
        		fittest = i;
        		record = fitnesses[i];
      		}
    	}
		//find the total fitness
		sum=0;
		for (i=0;i<pop_size;i++){
			sum+=fitnesses[i];
		}
		//now go back thru and update fitnesses as normalized
		for (i=0;i<pop_size;i++){
			if (fitnesses[i]>0) {
				temp = (fitnesses[i]/sum)*pop_size;
				fitnesses[i]=(char) temp;
			}
		}
		//update sum based on normalized fitness values
		sum=0;
		for (i=0;i<pop_size;i++){
			sum+=fitnesses[i];
		}
		//divvy up the percentages based on fitness
		value = ((32767)-pop_size)/sum; //this gives a slightly unfair advantage to the last member of the population
		//iterate thru pop and allocate points
		for (i=0;i<pop_size;i++){
			fitnesses[i]=((fitnesses[i]*value)+1);
			if (i>0) fitnesses[i]+=fitnesses[i-1];
		}
		//choose parents
		for (n=0;n<pop_size;n+=2){
			//choose a random number, look up where it falls
			value=rand();
			for (i=0;i<pop_size;i++){
				if (i==0){
					if (value<=fitnesses[i]) break;
				}
				else if ((value<=fitnesses[i])&&(value>fitnesses[i-1])) break;
			}
			//i is our first parent
			
			value=rand();
			for (j=0;j<pop_size;j++){
				if (j==0){
					if (value<=fitnesses[j]) break;
				}
				else if ((value<=fitnesses[j])&&(value>fitnesses[j-1])) break;
			}
			//j is our second parent
			//now that we have 2 breeders lets make 2 babies
			//choose a crossover point
			c_index = rand()/6534; //get a value b/t 0 and 5
			c_index2 = (rand()/6534)+5; //get a value b/t 6 and 10

			//give the first section of parent i to baby n
			for (l=0;l<c_index;l++){
				pop_new[n][l]=pop[i][l];
			}
			//give the second section of parent j to baby n
			for (l=c_index;l<c_index2;l++){
				pop_new[n][l]=pop[j][l];
			}
			//give the third section of parent i to baby n
			for (l=c_index2;l<dna_size;l++){
				pop_new[n][l]=pop[i][l];
			}
			
			//give the first section of parent j to baby n+1
			for (l=0;l<c_index;l++){
				pop_new[n+1][l]=pop[j][l];
			}
			//give the second section of parent i to baby n+1
			for (l=c_index;l<c_index2;l++){
				pop_new[n+1][l]=pop[i][l];
			}
			//give the third section of parent j to baby n+1
			for (l=c_index2;l<dna_size;l++){
				pop_new[n+1][l]=pop[j][l];
			}

			//mutation
			for (l = 0; l < dna_size; l++) {
				if ((rand()/327) < mutation_rate) {
					pop_new[n][l] = (rand()/1213)+95; //generate random mutation in baby1
				}
			}
			for (l = 0; l < dna_size; l++) {
				if ((rand()/327) < mutation_rate) {
					pop_new[n+1][l] = (rand()/1213)+95; //generate random mutation in baby2
				}
			}
		}
		//replace 1 member of new generation with fittest from old generation
		for (i=0;i<dna_size;i++){
			pop_new[0][i] = pop[fittest][i];
		}
		//make old population == new population
		for (i=0;i<pop_size;i++){
			for (j=0;j<dna_size;j++){
				pop[i][j] = pop_new[i][j];
			}
		}
	    //c. Evaluate the new population
		for (i=0;i<pop_size;i++){
			evaluate(i);
		}
   		
		record = 0;
    	for (i = 0; i < pop_size; i++) {
      		if (fitnesses[i] > record) {
        		c_index = i;
        		record = fitnesses[i];
      		}
    	}
		//display best
		WriteUSART(254);
		WriteUSART(128); //set cursor to 0 position
		for (b=0;b<dna_size;b++){
			putc(pop[c_index][b],stdout); //print ascii
			delay();
			delay();
		}
		//new line
		WriteUSART(254);
		WriteUSART(191);
		//write the score
		
		itoa(record, str);
		printf("fit:%s ", str);
		//write the iterations
		itoa(count, str);
		printf("time:%s", str);
		count++;
		//check for success
		for (i=0;i<pop_size;i++){
			if (fitnesses[i]>=goal){
				quit(); //enter infinite loop
			}
		}
	
	//decrease mutation rate as we get closer
	if ((record>5)&&(mutation_rate>5)) mutation_rate-=1;
	}//repeat until goal achieved
	return;
}